package simulator;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import MainLibrary.AbstractCollisionDomainStrategy;
import MainLibrary.AbstractDomain;
import MainLibrary.CarrierObj;
import MainLibrary.EventHandlerForReceptors;
import MainLibrary.HeapSorterReceptors;
import MainLibrary.LigandObserver;
import MainLibrary.MobilityModel;
import MainLibrary.UniversalReceptor;
import com.threed.jpct.SimpleVector;


public class CylindricReceptorDomain extends CylindricDomain{

	protected int receptorType;
	protected ArrayList<UniversalReceptor> receptorList=new ArrayList<UniversalReceptor>();
	protected ArrayList<LigandObserver> obsList=new ArrayList<LigandObserver>();
	protected Random rand=new Random();
	protected Map<Object,Object> oldExport=new HashMap<Object,Object>();	
	public EventHandlerForReceptors eventHandler=new EventHandlerForReceptors();

	public CylindricReceptorDomain(AbstractDomain parentDomain, MobilityModel model, AbstractCollisionDomainStrategy strategy,
			boolean visible, boolean export, float h, float r, float[] dCenter, int receptorType, int receptorQuantity, ArrayList<LigandObserver> ligandObserver,  boolean useOctree) {
		super(parentDomain, model, strategy, visible, export, h, r, dCenter, useOctree);
		this.receptorType=receptorType;
		this.obsList=ligandObserver;

		if(visibleOn3D && getManager().is3DEnable) getManager().engine.changeDomainColor(this, Color.green);		
		createReceptors(receptorQuantity);
	}

	public void handleBoundCollision(CarrierObj c){		
		if(c.getDomain()==this){
			getCollisionStrategy().relocateInnerObjs(c,this);
		}else{
			checkIncomingCarrier(c);
		}
	}

	public void checkIncomingCarrier(CarrierObj c){		
		boolean catched=false;
		float x,  y,  z,  r;

		float dt=(float) getCollisionStrategy().getExternalCollisionTime(c, this);	
		float[] velocity=c.getVelocityComponent();

		// Carrier Coords on the Cylindric External Surface
		x=c.getCenter()[0]-velocity[0]*dt;
		y=c.getCenter()[1]-velocity[1]*dt;
		z=c.getCenter()[2]-velocity[2]*dt;		

		r=(float) c.getRadius();
		float[] coords=null;				

		for (int i=0;i<receptorList.size() && !catched;i++){
			if((receptorList.get(i).isFree() && receptorList.get(i).canDownload())){
				coords=receptorList.get(i).getAbsoluteCoords();
				float dim=receptorList.get(i).radius;								
				float dx = (coords[0]-x), dy = (coords[1]-y), dz= (coords[2]-z);
				float distance =(float) Math.sqrt(dx*dx+dy*dy+dz*dz);									
				if (distance<((dim+r)) && (receptorList.get(i).checkCarrierCompatibility(c))){	
					System.out.println("== RECEPTOR DOWNLOAD!");
					catched=true;
					receptorList.get(i).configureForDownload(c);						
				}
			}
		}
		if(!catched){			
			getCollisionStrategy().relocateOuterObjs(c,this);
		}
	} 

	public ArrayList<UniversalReceptor> getReceptors(){
		return receptorList;
	}

	public void createReceptors(int n){					
		float theta=0;
		float fi=0;	
		int rad=(int) getRadius();

		float[] cdCenter1=new float[3];		
		int created=0;
		float height=(float)getHeight();	

		while(created<n){
			UniversalReceptor r = null;	

			float xx=(float)((rand.nextFloat()*2*rad)-rad);
			float zz=(float)((rand.nextFloat()*2*rad)-rad);

			// Relative Coords!
			cdCenter1[0]=xx;
			cdCenter1[1]=-height/2;
			cdCenter1[2]=zz;

			r=new CD40UniversalReceptor(this, parentDomain, true, cdCenter1, theta, fi, getManager().configWrapper.receptorStateConfig.get(receptorType+"-radius").shortValue(),receptorType, this.eventHandler);

			receptorList.add(r);
			receptorSetup(r); 
			created++;
		}		

		removeOverlappedReceptors(receptorList);

		// Write Receptors Table
		WriteReceptorsFile w=WriteReceptorsFile.getInstance();
		w.initialize();
		w.export(receptorList);
	}	

	public void removeOverlappedReceptors(ArrayList<UniversalReceptor> list){

		double radius=getRadius();
		float[] origin=new float[3];
		origin[0]=getAbsCenter()[0];
		origin[1]=getAbsCenter()[1];
		origin[2]=getAbsCenter()[2];

		SimpleVector coords1=new SimpleVector();
		double radius1=0;
		SimpleVector coords2=new SimpleVector();
		double radius2=0;	

		List<UniversalReceptor> receptorToRemove=new ArrayList<UniversalReceptor>();

		for (int i=0;i<list.size();i++){

			float[] tempCoords1=new float[3];
			tempCoords1[0]=list.get(i).getAbsoluteCoords()[0];
			tempCoords1[1]=list.get(i).getAbsoluteCoords()[1];
			tempCoords1[2]=list.get(i).getAbsoluteCoords()[2];

			radius1=list.get(i).getRadius();

			float dx=tempCoords1[0]-origin[0];
			float dy=tempCoords1[1]-origin[1];
			float dz=tempCoords1[2]-origin[2];

			float distance=(float) Math.sqrt(dx*dx+dy*dy+dz*dz);

			if (distance+radius1>radius){				
				if(!(receptorToRemove.contains(list.get(i)))){
					receptorToRemove.add(list.get(i));									
				}				
			}
		}
		list.removeAll(receptorToRemove);
		for(int k=0;k<receptorToRemove.size();k++){			
			receptorToRemove.get(k).finalize();
		}	

		receptorToRemove.clear();	
		receptorToRemove=new ArrayList<UniversalReceptor>();		
		boolean isFinished=false;

		// Reorder first!!				
		HeapSorterReceptors h=new HeapSorterReceptors();
		h.sort(list);		
		// Reorder done...
		
		for (int i=0;i<list.size()-1;i++){
			if(receptorToRemove.contains(list.get(i))){
				isFinished=true;
			}else{
				isFinished=false;
			}

			for (int j=i+1;j<list.size() && !isFinished;j++){						
				if(list.get(j).getRelativeDistancePlusRadius()>=list.get(i).getRelativeDistanceMinusRadius()){					
					coords1=new SimpleVector(list.get(i).getAbsoluteCoords());
					radius1=list.get(i).getRadius();
					coords2=new SimpleVector(list.get(j).getAbsoluteCoords());
					radius2=list.get(j).getRadius();

					float dx=coords1.x-coords2.x;
					float dy=coords1.y-coords2.y;
					float dz=coords1.z-coords2.z;
					float distance=(float) Math.sqrt(dx*dx+dy*dy+dz*dz);

					if (distance<(radius1+radius2)){

						if(!(receptorToRemove.contains(list.get(j)))){
							receptorToRemove.add(list.get(j));														
						}						
					}					
				}else{
					isFinished=true;
				}			
			}
		}		

		list.removeAll(receptorToRemove);
		for(int k=0;k<receptorToRemove.size();k++){					
			receptorToRemove.get(k).finalize();	
		}	

		receptorToRemove.clear();
		receptorToRemove=null;
	}	

	public void receptorSetup(UniversalReceptor r) {
		List<LigandObserver> obsList=getObsList();
		int index=rand.nextInt(obsList.size());
		r.attachObs(obsList.get(index));
	}

	public List<LigandObserver> getObsList(){
		return obsList;
	}

	@Override
	public Map<Object, Object> export() {
		exportString=new HashMap<Object, Object>();		
		int currentID=0;
		int assimilations=0;

		for(int i=0;i<receptorList.size();i++){		
			currentID=((CD40UniversalReceptor) receptorList.get(i)).getID();
			assimilations=((CD40UniversalReceptor) receptorList.get(i)).assimilatedInfo;
			exportString.put("Receptor-"+i, currentID);	
			exportString.put("Assimilations-"+i, assimilations);	
		}
		
		if(oldExport.equals(exportString)){
			exportString=null;
		}else{
			oldExport=exportString;
		}
		return exportString;
	}

	@Override
	public void trigger() {		
		for (int i=0;i<receptorList.size();i++){			
			receptorList.get(i).trigger();			
		}				
	}
}
