package simulator;

import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import MainLibrary.AbstractCollisionDomainStrategy;
import MainLibrary.AbstractDomain;
import MainLibrary.AbstractThreadingStrategy;
import MainLibrary.CheckNanoObjCollisionThreadLatch;
import MainLibrary.HeapSorter;
import MainLibrary.MobilityModel;
import MainLibrary.NanoObj;


public class NewDiffusionThreadingStrategy extends	AbstractThreadingStrategy {

	private final int maxThreads = Runtime.getRuntime().availableProcessors();
	private ExecutorService executorThreads = Executors.newFixedThreadPool(maxThreads);	 

	public NewDiffusionThreadingStrategy(MobilityModel m) {
		super(m);
	}

	/* (non-Javadoc)
	 * @see MainLibrary.AbstractThreadingMoveStrategy#move(java.util.List)
	 */
	@Override
	public void move(List<NanoObj> nanoList) {
		int div=maxThreads;
		int size=nanoList.size();		
		CountDownLatch latch=new CountDownLatch(1);

		if(size>div){
			int intPart=size/div;	

			if (intPart>0) {
				latch=new CountDownLatch(div);
			}				

			for(int i=0;i<div-1;i++){
				executorThreads.execute(new DiffusionModelMoveThread(nanoList.subList(intPart*i, intPart*(i+1)),latch));			
			}
			executorThreads.execute(new DiffusionModelMoveThread(nanoList.subList(intPart*(div-1), size),latch));

		}else if(size>0){
			executorThreads.execute(new DiffusionModelMoveThread(nanoList,latch));
		}		

		try {
			latch.await();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}		
	}

	public void checkCollision(List<NanoObj> objsToCheck) {

		AbstractDomain activeDomain=null;
		if (objsToCheck.size()>0){
			activeDomain=objsToCheck.get(0).getDomain();			

			
			// Reorder first!!		
			HeapSorter h=new HeapSorter();
			h.sort(objsToCheck);	
			


			// Check Collision!!			
			try {				
				int threadCount=maxThreads*4;
				int subList=objsToCheck.size()/threadCount;
				int residue=objsToCheck.size()%threadCount;

				CountDownLatch latch=new CountDownLatch(1);
				if (subList>0) {
					latch=new CountDownLatch(threadCount);
				}
				if (subList>0) {
					for (int i=0;i<threadCount-1;i++){	
						executorThreads.execute(new CheckNanoObjCollisionThreadLatch(i*subList, (i+1)*subList,objsToCheck,m,activeDomain,latch));								
					}
				}			
				if (residue==0) {	
					executorThreads.execute(new CheckNanoObjCollisionThreadLatch((threadCount-1)*subList, (threadCount)*subList,objsToCheck,m,activeDomain,latch));			
				} else {
					executorThreads.execute(new CheckNanoObjCollisionThreadLatch((threadCount-1)*subList, objsToCheck.size(),objsToCheck,m,activeDomain,latch));
				}

				latch.await();	
				
				activeDomain.joinAllCollisionsLists();
				

			} catch (Exception e1) {				
				e1.printStackTrace();
			}	


			//======================================================
			//     FIXME STABLE but without MultiThreading!!! 
			//======================================================
			try {
				// Check the collision with other domains!
				if (objsToCheck.size()>0){	
					int objSize=objsToCheck.size();

					List<AbstractDomain> domainToCheck=m.reorderDomain_byHeap(activeDomain);

					//Check Collision!!
					boolean isFinish=false;

					for (int i=0;i<domainToCheck.size();i++){
						isFinish=false;
						try{
							for (int j=0;j<objSize && !isFinish;j++){
								if(objsToCheck.get(j).getDomain()==activeDomain){									
									float Dplus=domainToCheck.get(i).getDistance()+domainToCheck.get(i).getBoundingRadius();
									float Dminus=domainToCheck.get(i).getDistance()-domainToCheck.get(i).getBoundingRadius();
									float Nplus=objsToCheck.get(j).getAbsDistancePlusRadius();
									float Nminus=objsToCheck.get(j).getAbsDistanceMinusRadius();									

									if(Dminus>Nplus) { 										
										// They are not overlapped, neither the following objects!	
										isFinish=true;
										
									} else if((Dminus<=Nminus && Nminus<=Dplus) || (Dminus>=Nminus && Nplus>=Dminus) || (Dplus<=Nplus && Nminus<=Dplus)) { 
										// probable collision!						
										try{
											domainToCheck.get(i).getCollisionStrategy().checkExternalCollisionBound(domainToCheck.get(i), objsToCheck.get(j));

										}catch (Exception e){										
											e.printStackTrace();						
										}
									} 
								}
							}
						}catch (Exception e){
							e.printStackTrace();
						}
					}
				}
			} catch (Exception e1) {							
				e1.printStackTrace();
			}
		}		
	}	

}
